/*
 * SPC device class.
 *
 * Copyright (C) 2023 by NTC MacroSystems LTD
 */

#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/vmalloc.h>
#include <linux/poll.h>
#include <linux/idr.h>
#include <linux/mm.h>
#include <linux/delay.h>
#include <linux/sched.h>
#include <linux/pci.h>
#include <linux/time.h>
#include <asm/div64.h>

#include "spc_device.h"

MODULE_AUTHOR("Semen M. Krylov <ksm@ntcms.ru>");
MODULE_DESCRIPTION("SPC device class");
MODULE_LICENSE("GPL");

#define DRIVER_NAME "spc"

#define SPC_MAX_DEVICES	16
static struct class *spc_class;
static dev_t spc_devt;
static DEFINE_IDA(ida);
static DEFINE_MUTEX(ida_lock);

static inline struct spc_device *device_to_spc_device(struct device *dev)
{
	return container_of(dev, struct spc_device, dev);
}

static inline struct spc_device *inode_to_spc_device(struct inode *inode)
{
	return container_of(inode->i_cdev, struct spc_device, cdev);
}

static int spc_open(struct inode *inode, struct file *filp)
{
	struct spc_device *dev = inode_to_spc_device(inode);
	struct spc_fd *fd;
	unsigned long flags;
	int ret;

	/* Just check if there is already an fd object present.
	   If so there is no need even to allocate one.
	   'cos we will probably fail to acquire this device.
	 */
	spin_lock_irqsave(&dev->lock, flags);
	if (dev->fd) {
		spin_unlock_irqrestore(&dev->lock, flags);
		return -EBUSY;
	}
	spin_unlock_irqrestore(&dev->lock, flags);

	/* Ok. It seems to be no owner.
	   Allocate descriptor. */
	fd = kmalloc(sizeof(*fd), GFP_KERNEL);
	if (!fd)
		return -ENOMEM;

	fd->dev = dev;
	fd->device_mmaped = 0;

	/* Now. Let's try to capture the device. */
	spin_lock_irqsave(&dev->lock, flags);
	if (dev->fd) {
		spin_unlock_irqrestore(&dev->lock, flags);
		kfree(fd);
		return -EBUSY;
	}
	/* Succeded. No we have exclusive access to
	   this device.*/
	filp->private_data = fd;
	spin_unlock_irqrestore(&dev->lock, flags);

	if (dev->ops->open) {
		ret = dev->ops->open(dev);
		if (ret) {
			/* And it failed. */
			dev->fd = 0;
			kfree(fd);
			return ret;
		}
	}

	return 0;
}

static int spc_release(struct inode *inode, struct file *filp)
{
	struct spc_fd *fd = filp->private_data;
	struct spc_device *dev = fd->dev;

	dev->fd = 0;

	if (fd->device_mmaped) {
		fd->device_mmaped = 0;
		dev->ops->munmap(dev, fd);
	}

	kfree(fd);

	if (dev->ops->close)
		dev->ops->close(dev);

	return 0;
}

static int spc_mmap(struct file *filp, struct vm_area_struct *vma)
{
	struct spc_fd *fd = filp->private_data;
	struct spc_device *dev = fd->dev;
	unsigned long off, len;
	int i;
	int res;

	if (vma->vm_pgoff >= (~0ul >> PAGE_SHIFT)) {
		printk(KERN_WARNING "spc: mmap offset out of range\n");
		return -EINVAL;
	}

	off = vma->vm_pgoff << PAGE_SHIFT;
	len = PAGE_ALIGN(((unsigned long)dev->mem & PAGE_MASK) +
			sizeof(*dev->mem));

	if (!dev->ops->mmap) {
		printk(KERN_WARNING "spc: no device specific mmap\n");
		return -EINVAL;
	}

	/* Find requested device private memory region. */
	for (i = 0; i < dev->caps->pmr_nr; i++)
		if (dev->caps->pmr_offs[i] == off)
			break;
	if (i == dev->caps->pmr_nr) {
		printk(KERN_WARNING "spc: device has no private"
			       " memory region with offset 0x%08x\n",
			       (unsigned int)off);
		return -EINVAL;
	}
	res = dev->ops->mmap(dev, vma);
	if (!res) {
		fd->device_mmaped = 1;
		fd->device_mmap_addr = (volatile void __iomem *)vma->vm_start;
	}

	return res;
}

static int ioctl_get_caps(struct spc_fd *fd,
		struct spc_device *dev,
		struct spc_caps __user *p)
{
	if (copy_to_user(p, dev->caps, sizeof(*dev->caps)))
		return -EFAULT;
	return 0;
}

static int ioctl_get_rev(struct spc_fd *fd,
		struct spc_device *dev,
		int __user *p)
{
	if (copy_to_user(p, &dev->rev, sizeof(dev->rev)))
		return -EFAULT;
	return 0;
}

static int ioctl_get_hid(struct spc_fd *fd,
		struct spc_device *dev,
		struct spc_hid __user *p)
{
	if (copy_to_user(p, &dev->hid, sizeof(dev->hid)))
		return -EFAULT;
	return 0;
}

static int ioctl_check_reset(struct spc_fd *fd,
		struct spc_device *dev,
		unsigned int __user *status)
{
	return dev->ops->check_reset ? dev->ops->check_reset(dev, status) : -ENOSYS;
}

static long spc_ioctl(struct file *filp, unsigned int cmd,
		unsigned long __user arg)
{
	struct spc_fd *fd = filp->private_data;
	struct spc_device *dev = fd->dev;
	void __user *p = (void __user *)arg;

	switch (cmd) {
	case SPC_IOCTL_GET_CAPS:
		return ioctl_get_caps(fd, dev, p);
	case SPC_IOCTL_GET_REV:
		return ioctl_get_rev(fd, dev, p);
	case SPC_IOCTL_GET_HID:
		return ioctl_get_hid(fd, dev, p);
	case SPC_IOCTL_CHECK_RESET:
		return ioctl_check_reset(fd, dev, p);
	default:
		if (dev->ops->ioctl)
			return dev->ops->ioctl(fd, dev, cmd, p);
		else
			break;
	};

	return -EINVAL;
}

static struct file_operations spc_fops = {
	.owner = THIS_MODULE,
	.open = spc_open,
	.release = spc_release,
	.mmap = spc_mmap,
	.unlocked_ioctl = spc_ioctl,
};

struct attribute attributes[] = {
	{.name = __stringify(name), .mode = VERIFY_OCTAL_PERMISSIONS(S_IRUGO)},
	{.name = __stringify(caps), .mode = VERIFY_OCTAL_PERMISSIONS(S_IRUGO)},
	{}
};
static const struct attribute_group spc_attrs_group[] = {
	{.attrs = (struct attribute **)&attributes,},
};

static void spc_destruct(struct device *d)
{
	struct spc_device *dev = device_to_spc_device(d);

	mutex_lock(&ida_lock);
	ida_free(&ida, dev->id);
	mutex_unlock(&ida_lock);

	if (dev->ops->release)
		dev->ops->release(dev);
	if (dev->mem) {
		vfree(dev->mem);
		dev->mem = 0;
	}
}

int spc_register(struct spc_device *dev)
{
	int ret;

	BUG_ON(!dev);

	BUG_ON(dev->caps->name[0] == 0 ||
	       dev->caps->name[sizeof(dev->caps->name)-1] != 0);

	BUG_ON(dev->ops == 0);
	BUG_ON(dev->ops->release == 0);

	mutex_lock(&ida_lock);
	dev->id = ida_alloc(&ida, GFP_KERNEL);
	ret = (dev->id < 0 ) ? dev->id : 0;
	mutex_unlock(&ida_lock);

	if (ret) {
		printk(KERN_INFO DRIVER_NAME ": ida_alloc failed %d\n", ret);
		goto out_idr;
	}

	dev->id &= ((1 << IDA_BITMAP_LONGS) - 1);
	if (dev->id >= SPC_MAX_DEVICES) {
		printk(KERN_INFO DRIVER_NAME ": too big dev id %d vs %d\n", dev->id, SPC_MAX_DEVICES);
		ret = -EBUSY;
		goto out_idr;
	}

	spin_lock_init(&dev->lock);
	dev->fd = 0;

	dev->mem = dma_alloc_coherent(&(dev->pcid->dev), sizeof(*dev->mem),
					&dev->pmem, GFP_ATOMIC);
	dev->caps->shmem_offs = dev->pmem;
	dev->caps->shmem_size = sizeof(*dev->mem);

	if (!dev->mem) {
		printk(KERN_INFO DRIVER_NAME ": memory allocion failed\n");
		ret = -ENOMEM;
		goto out_idr;
	}

	dev->dev.class = spc_class;
	dev->dev.release = spc_destruct;
	dev_set_name(&dev->dev, "spc-%d", dev->id);
	dev->dev.devt = MKDEV(MAJOR(spc_devt), dev->id);

	cdev_init(&dev->cdev, &spc_fops);
	dev->cdev.owner = dev->owner;
	ret = cdev_add(&dev->cdev, dev->dev.devt, 1);
	if (ret < 0) {
		printk(KERN_ERR "spc: failed to create a char dev\n");
		goto out_mem;
	}

	printk(KERN_INFO "spc: registered new device \"%s\", HID \"%s\"\n",
		dev->caps->name, dev->hid.hid);

	return 0;

out_mem:
	dma_free_coherent(&(dev->pcid->dev), sizeof(*dev->mem),
				dev->mem, dev->pmem);
	dev->mem = 0;
out_idr:
	mutex_lock(&ida_lock);
	ida_free(&ida, dev->id);
	mutex_unlock(&ida_lock);

	return ret;
}
EXPORT_SYMBOL(spc_register);

void spc_unregister(struct spc_device *dev)
{
	cdev_del(&dev->cdev);
	dma_free_coherent(&(dev->pcid->dev), sizeof(*dev->mem),
				dev->mem, dev->pmem);
}
EXPORT_SYMBOL(spc_unregister);

static int __init spc_init(void)
{
	int ret;

	spc_class = class_create(THIS_MODULE, "spc");
	if (IS_ERR(spc_class)) {
		printk(KERN_ERR "spc: class_create() failed\n");
		return PTR_ERR(spc_class);
	}

	ret = alloc_chrdev_region(&spc_devt, 0, SPC_MAX_DEVICES, "spc");
	if (ret < 0) {
		printk(KERN_ERR "spc: register_chrdev() failed\n");
		goto free_class;
	}

#if defined(CONFIG_SYSFS)
	spc_class->dev_groups = (const struct attribute_group **)&spc_attrs_group;
#endif

	return 0;
free_class:
	class_destroy(spc_class);

	return ret;
}

static void __exit spc_exit(void)
{
	unregister_chrdev_region(spc_devt, SPC_MAX_DEVICES);
	class_destroy(spc_class);
}

subsys_initcall(spc_init);
module_exit(spc_exit);
